<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace ServicenowModule\Command;

use App\Model\Table\ProxiesTable;
use App\Model\Table\SystemsettingsTable;
use Cake\Console\Arguments;
use Cake\Console\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use ErrorException;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;
use itnovum\openITCOCKPIT\Exceptions\HostNotFoundException;
use itnovum\openITCOCKPIT\Exceptions\ServiceNotFoundException;
use ServicenowModule\Lib\ServicenowData;
use ServicenowModule\Model\Table\ServicenowHostspecificSettingsTable;
use ServicenowModule\Model\Table\ServicenowIncidentsTable;
use ServicenowModule\Model\Table\ServicenowServicespecificSettingsTable;
use ServicenowModule\Model\Table\ServicenowSettingsTable;

/**
 * Class ServicenowNotificationCommand
 * @package ServicenowModule\Command
 */
class ServicenowNotificationCommand extends Command {

    private $servicenowSettings;
    private $servicenowSpecificSettings;
    private $proxySettings;
    private $incident;
    private $create_new_incident = false;
    private $send_notification = true;

    /**
     * @param ConsoleOptionParser $parser
     * @return ConsoleOptionParser
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        /** @var ConsoleOptionParser $parser */
        $parser = parent::buildOptionParser($parser);

        $parser->addOptions([
            'type'         => ['short' => 't', 'help' => __d('oitc_console', 'Type of the notification host or service')],
            'hostname'     => ['help' => __d('oitc_console', 'Host uuid you want to send a notification')],
            'hoststate'    => ['help' => __d('oitc_console', 'current host state')],
            'hostaddress'  => ['help' => __d('oitc_console', 'host address')],
            'output'       => ['help' => __d('oitc_console', 'output')],
            'servicedesc'  => ['help' => __d('oitc_console', 'Service uuid you want to notify')],
            'servicestate' => ['help' => __d('oitc_console', 'service state')],
            'comment'      => ['help' => __d('oitc_console', 'comment')],
        ]);

        return $parser;
    }

    /**
     * @param Arguments $args
     * @param ConsoleIo $io
     * @throws HostNotFoundException
     * @throws ServiceNotFoundException
     * @throws ErrorException
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        /** @var ServicenowSettingsTable $ServicenowSettingsTable */
        $ServicenowSettingsTable = TableRegistry::getTableLocator()->get('ServicenowModule.ServicenowSettings');
        /** @var ServicenowIncidentsTable $ServicenowIncidentsTable */
        $ServicenowIncidentsTable = TableRegistry::getTableLocator()->get('ServicenowModule.ServicenowIncidents');
        /** @var ServicenowHostspecificSettingsTable $ServicenowHostspecificSettingsTable */
        $ServicenowHostspecificSettingsTable = TableRegistry::getTableLocator()->get('ServicenowModule.ServicenowHostspecificSettings');
        /** @var ServicenowServicespecificSettingsTable $ServicenowServicespecificSettingsTable */
        $ServicenowServicespecificSettingsTable = TableRegistry::getTableLocator()->get('ServicenowModule.ServicenowServicespecificSettings');
        /** @var ProxiesTable $ProxiesTable */
        $ProxiesTable = TableRegistry::getTableLocator()->get('Proxies');
        /** @var SystemsettingsTable $SystemsettingsTable */
        $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        /** @var ServicenowData $ServicenowData */
        $ServicenowData = new ServicenowData();

        $this->proxySettings = $ProxiesTable->getSettings();
        $this->servicenowSettings = $ServicenowSettingsTable->find()->first()->toArray();

        if ($args->getOption('type') !== '') {
            if ($args->getOption('type') === 'Host') {
                $notification = [
                    'host_uuid'    => $args->getOption('hostname'),
                    'host_state'   => (int)$args->getOption('hoststate'),
                    'host_address' => $args->getOption('hostaddress'),
                    'type'         => $args->getOption('type'),
                    'output'       => $args->getOption('output'),
                    'comment'      => $args->getOption('comment'),
                ];
            }

            if ($args->getOption('type') === 'Service') {
                $notification = [
                    'host_uuid'     => $args->getOption('hostname'),
                    'host_state'    => (int)$args->getOption('hoststate'),
                    'host_address'  => $args->getOption('hostaddress'),
                    'service_uuid'  => $args->getOption('servicedesc'),
                    'service_state' => $args->getOption('servicestate'),
                    'type'          => $args->getOption('type'),
                    'output'        => $args->getOption('output'),
                    'comment'       => $args->getOption('comment'),
                ];
            }
        }

        $systemAddress = $SystemsettingsTable->find()
            ->where([
                'key' => 'SYSTEM.ADDRESS'
            ])->first()->toArray()['value'];

        if ($notification['type'] == "Host") {
            $incidentEntity = $ServicenowIncidentsTable->find()
                ->where([
                    'ServicenowIncidents.host_uuid' => $notification['host_uuid']
                ])
                ->first();

            if (empty($incidentEntity)) {
                $this->incident = [];
            } else {
                $this->incident = $incidentEntity->toArray();
            }

            $hostSpecificSettings = $ServicenowHostspecificSettingsTable->find()
                ->where([
                    'ServicenowHostspecificSettings.host_uuid' => $notification['host_uuid']
                ])
                ->first();

            if (!empty($hostSpecificSettings)) {
                $this->servicenowSpecificSettings = Hash::merge($this->servicenowSettings, array_filter($hostSpecificSettings->toArray()));
            } else {
                $this->servicenowSpecificSettings = $this->servicenowSettings;
            }

            $host = $ServicenowData->getHost($notification['host_uuid']);

            if ($notification['host_state'] == "0")
                $state = "UP";
            else if ($notification['host_state'] == "1")
                $state = "DOWN";
            else
                $state = "UNREACHABLE";

            /*if (isset($host['tags'])) {
                $tags = str_replace(',', ' ', $host['tags']);
            }*/

            $hosturl = sprintf('https://%s/#!/hosts/browser/%s', $systemAddress, $host['id']);
            $serviceurl = '';
            $comment = 'URL: [code]<a href="' . $hosturl . '">' . $hosturl . '</a>[/code]';

            $short_description = sprintf("Host %s is %s", $host['name'], $state);

            $description = sprintf("Time: %s\n", date('H:i:s - d.m.Y T'));
            $description .= sprintf("Host Name: %s\n", $host['name']);
            $description .= sprintf("Hostaddress: %s\n", $notification['host_address']);
            $description .= sprintf("State: %s\n", $state);
            $description .= sprintf("Output: %s\n", $notification['output']);
        }

        if ($notification['type'] == "Service") {
            $incidentEntity = $ServicenowIncidentsTable->find()
                ->where([
                    'ServicenowIncidents.service_uuid' => $notification['service_uuid']
                ])
                ->first();

            if (empty($incidentEntity)) {
                $this->incident = [];
            } else {
                $this->incident = $incidentEntity->toArray();
            }

            $serviceSpecificSettings = $ServicenowServicespecificSettingsTable->find()
                ->where([
                    'ServicenowServicespecificSettings.service_uuid' => $notification['service_uuid']
                ])
                ->first();

            if (!empty($serviceSpecificSettings)) {
                $this->servicenowSpecificSettings = Hash::merge($this->servicenowSettings, array_filter($serviceSpecificSettings->toArray()));
            } else {
                $this->servicenowSpecificSettings = $this->servicenowSettings;
            }

            $host = $ServicenowData->getHost($notification['host_uuid']);
            $service = $ServicenowData->getService($notification['service_uuid']);

            /*if (isset($service['tags'])) {
                $tags = str_replace(',', ' ', $service['tags']);
            }*/

            $hosturl = sprintf('https://%s/#!/hosts/browser/%s', $systemAddress, $host['id']);
            $serviceurl = sprintf('https://%s/#!/services/browser/%s', $systemAddress, $service['id']);
            $comment = 'Host URL: [code]<a href="' . $hosturl . '">' . $hosturl . '</a><br>[/code]';
            $comment .= 'Service URL: [code]<a href="' . $serviceurl . '">' . $serviceurl . '</a>[/code]';

            if ($notification['service_state'] == "0")
                $state = "OK";
            else if ($notification['service_state'] == "1")
                $state = "WARNING";
            else if ($notification['service_state'] == "2")
                $state = "CRITICAL";
            else
                $state = "UNKNOWN";

            $short_description = sprintf("Service %s on Host %s is %s", $service['name'], $host['name'], $state);

            $description = sprintf("Time: %s\n", date('H:i:s - d.m.Y T'));
            $description .= sprintf("Service Name: %s\n", $service['name']);
            $description .= sprintf("Hostaddress: %s\n", $notification['host_address']);
            $description .= sprintf("State: %s\n", $state);
            $description .= sprintf("Output: %s\n", $notification['output']);
        }


        if ($state == "UP" || $state == "OK") {
            $incident_active_state = 'false';
            $incident_state = $this->servicenowSettings['recovery_state'];
        }
        if ($state == "CRITICAL" || $state == "DOWN") {
            $incident_active_state = 'true';
            $incident_state = '1';
        }
        if ($state == "WARNING") {
            $incident_active_state = 'true';
            $incident_state = '1';
        }
        if ($state == "UNKNOWN" || $state == "UNREACHABLE") {
            $incident_active_state = 'true';
            $incident_state = '1';
        }

        if (empty($this->incident)) {  //create new incident
            $this->create_new_incident = true;
            $data = [
                'short_description' => $short_description,
                'description'       => $description,
                'comments'          => $comment,
                'active'            => $incident_active_state,
                'category'          => 'openITCOCKPIT',
                'urgency'           => $this->servicenowSpecificSettings['urgency'], //see: https://docs.servicenow.com/bundle/london-it-service-management/page/product/incident-management/reference/r_PrioritizationOfIncidents.html?cshalt=yes
                'impact'            => $this->servicenowSpecificSettings['impact'],
                //'knowledge' => 'true',
                'u_host_url'        => $hosturl,
                'u_service_url'     => $serviceurl,
                //'sys_tags'          => $tags,
                'caller_id'         => $this->servicenowSettings['user_id']
            ];
        } else {    //update existing incident
            $data = [
                'short_description' => $short_description,
                'description'       => $description,
                'active'            => $incident_active_state,
                'state'             => $incident_state, //1 = "New"; 2 = "In Progress"; 3 = "On Hold"; 4 = ""; 5 = ""; 6 = "Resolved"; 7 = "Closed"; 8 = "Canceled";
                'category'          => 'openITCOCKPIT',
                'urgency'           => $this->servicenowSpecificSettings['urgency'], //see: https://docs.servicenow.com/bundle/london-it-service-management/page/product/incident-management/reference/r_PrioritizationOfIncidents.html?cshalt=yes
                'impact'            => $this->servicenowSpecificSettings['impact'],
                'caller_id'         => $this->servicenowSettings['user_id']
            ];
        }

        if (isset($notification['comment']) && !empty($notification['comment']) && !$this->create_new_incident) {
            $data['comments'] = $notification['comment'];
        }

        if ($state == "UP" || $state == "OK") {
            $data['close_code'] = 'Closed/Resolved by Caller';
            $data['close_notes'] = $notification['output'];

            if ($this->create_new_incident) {
                $this->send_notification = false;
            }
        }

        if (!empty($this->servicenowSpecificSettings['assignment_group'])) {
            $data['assignment_group'] = $this->servicenowSpecificSettings['assignment_group'];
        }
        if (!empty($this->servicenowSpecificSettings['assigned_to'])) {
            $data['assigned_to'] = $this->servicenowSpecificSettings['assigned_to'];
        }
        if (!empty($this->servicenowSpecificSettings['business_service'])) {
            $data['business_service'] = $this->servicenowSpecificSettings['business_service'];
        }


        $options = [
            'json'    => $data,
            'auth'    => [$this->servicenowSettings['user_id'], $this->servicenowSettings['user_password']],
            'headers' => [
                'Accept'       => 'application/json',
                'Content-Type' => 'application/json'
            ]
        ];

        if ($this->servicenowSettings['use_proxy']) {
            $options['proxy'] = [
                'http'  => sprintf('%s:%s', $this->proxySettings['ipaddress'], $this->proxySettings['port']),
                'https' => sprintf('%s:%s', $this->proxySettings['ipaddress'], $this->proxySettings['port'])
            ];
        } else {
            $options['proxy'] = [
                'http'  => false,
                'https' => false
            ];
        }

        $guzzleclient = new Client();
        $serviceNowUrl = sprintf('%s/api/now/table/incident', $this->servicenowSettings['domain']);

        if ($this->send_notification) {
            try {
                if ($this->create_new_incident) {
                    $response = $guzzleclient->post($serviceNowUrl, $options);
                } else {
                    if ($incident_active_state == 'true' || ($incident_active_state == 'false' && $this->incident['state'] != 7)) {
                        try {
                            $response = $guzzleclient->patch($serviceNowUrl . '/' . $this->incident['sys_id'], $options);
                        } catch (RequestException $e) {
                            echo $e->getMessage();
                        }
                    }
                }

                if (isset($response)) {
                    $result = json_decode($response->getBody()->getContents(), true)['result'];
                    $incident_id = $result['number'];
                    $sys_id = $result['sys_id'];
                }

                if ($incident_active_state == 'true' && isset($response) && $incident_id != null) {   //insert or update incident entry
                    if ($this->create_new_incident) { //insert
                        $this->incident = [
                            'incident_id'  => $incident_id,
                            'sys_id'       => $sys_id,
                            'host_uuid'    => $notification['host_uuid'],
                            'service_uuid' => isset($notification['service_uuid']) ? $notification['service_uuid'] : '',
                        ];
                        $incidentEntity = $ServicenowIncidentsTable->newEntity($this->incident);
                        $ServicenowIncidentsTable->save($incidentEntity);
                        if ($incidentEntity->hasErrors()) {
                            throw new ErrorException();
                        }
                    } else {    //update
                        $this->incident['state'] = intval($incident_state);
                        $this->incident['incident_id'] = $incident_id;
                        $this->incident['sys_id'] = $sys_id;
                        $this->incident['modified'] = date("Y-m-d H:i:s");

                        if (isset($this->incident['service_uuid'])) {
                            $incidentEntity = $ServicenowIncidentsTable->find()
                                ->where([
                                    'ServicenowIncidents.service_uuid' => $this->incident['service_uuid']
                                ])
                                ->first();
                        } else {
                            $incidentEntity = $ServicenowIncidentsTable->find()
                                ->where([
                                    'ServicenowIncidents.host_uuid' => $this->incident['host_uuid']
                                ])
                                ->first();
                        }

                        $incidentEntity = $ServicenowIncidentsTable->patchEntity($incidentEntity, $this->incident);
                        $ServicenowIncidentsTable->save($incidentEntity);
                        if ($incidentEntity->hasErrors()) {
                            throw new ErrorException();
                        }
                    }
                } else if ($incident_active_state != 'true') {    //delete incident entry
                    $incidentEntity = $ServicenowIncidentsTable->get($this->incident['id']);
                    $ServicenowIncidentsTable->delete($incidentEntity);
                }

            } catch (RequestException $e) {
                echo $e->getMessage();
            }
        }

    }
}
