<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace ServicenowModule\Model\Table;

use Cake\Datasource\EntityInterface;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use ServicenowModule\Model\Entity\ServicenowSetting;

/**
 * ServicenowSettings Model
 *
 * @method ServicenowSetting get($primaryKey, $options = [])
 * @method ServicenowSetting newEntity($data = null, array $options = [])
 * @method ServicenowSetting[] newEntities(array $data, array $options = [])
 * @method ServicenowSetting|false save(EntityInterface $entity, $options = [])
 * @method ServicenowSetting saveOrFail(EntityInterface $entity, $options = [])
 * @method ServicenowSetting patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method ServicenowSetting[] patchEntities($entities, array $data, array $options = [])
 * @method ServicenowSetting findOrCreate($search, callable $callback = null, $options = [])
 *
 * @mixin TimestampBehavior
 */
class ServicenowSettingsTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('servicenow_settings');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->nonNegativeInteger('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->scalar('domain')
            ->maxLength('domain', 255)
            ->urlWithProtocol('domain')
            ->notEmptyString('domain');

        $validator
            ->scalar('user_id')
            ->maxLength('user_id', 255)
            ->notEmptyString('user_id');

        $validator
            ->scalar('user_password')
            ->maxLength('user_password', 255)
            ->notEmptyString('user_password');

        $validator
            ->scalar('assignment_group')
            ->maxLength('assignment_group', 255)
            ->allowEmptyString('assignment_group');

        $validator
            ->scalar('assigned_to')
            ->maxLength('assigned_to', 255)
            ->allowEmptyString('assigned_to');

        $validator
            ->scalar('business_service')
            ->maxLength('business_service', 255)
            ->allowEmptyString('business_service');

        $validator
            ->notEmptyString('impact')
            ->greaterThan('impact', 0);

        $validator
            ->notEmptyString('urgency')
            ->greaterThan('urgency', 0);

        $validator
            ->notEmptyString('recovery_state')
            ->greaterThan('recovery_state', 0);


        $validator
            ->boolean('two_way')
            ->notEmptyString('two_way');

        $validator
            ->boolean('use_proxy')
            ->notEmptyString('use_proxy');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        return $rules;
    }

    /**
     * @return array|EntityInterface|ServicenowSetting|null
     */

    /**
     * @return array|EntityInterface
     */
    public function getServicenowSettings() {
        try {
            $result = $this->find()
                ->firstOrFail();
            return $result;
        } catch (RecordNotFoundException $e) {
            return [
                'id'               => $this->getConfigurationId(), //its 1 every time
                'domain'           => null,
                'user_id'          => null,
                'user_password'    => null,
                'assignment_group' => null,
                'assigned_to'      => null,
                'business_service' => null,
                'impact'           => 1,
                'urgency'          => 1,
                'recovery_state'   => 6,
                'two_way'          => false,
                'use_proxy'        => false
            ];
        }
    }

    /**
     * @return int 1
     */
    public function getConfigurationId() {
        // 1 is the default id of the servicenow configuration because at the moment
        // openITCOCKPIT supports only one servicenow configuration.
        return 1;
    }

    /**
     * @return array|EntityInterface
     */
    public function getServicenowSettingsEntity() {
        try {
            return $this->find()->firstOrFail();
        } catch (RecordNotFoundException $e) {
            return $this->newEmptyEntity();
        }
    }
}
